import torch
import os
import sys
import pandas as pd
import numpy as np
import random

import model_utils as mu
from PIL import Image

sys.path.append('./utils')
BASE_DIR = "."

from torch.utils import data
from torchvision import transforms
from copy import deepcopy
from tqdm import tqdm


toTensor = transforms.ToTensor()
toPILImage = transforms.ToPILImage()

def pil_loader(path):
    img = Image.open(path)
    return img.convert('RGB')


# Constant factor which converts from a 30fps video to the corresponding spectrogram
imgFrameIdxToAudioIdxFactor = 86.15 / 30.0


def get_spectrogram_window_length(seq_len, num_seq, downsample):
    return int(imgFrameIdxToAudioIdxFactor * seq_len * num_seq * downsample)


class HierarchicalHomage(data.Dataset):

    def __init__(
        self,
        mode='train',
        transform=None,
        seq_len=5,
        num_seq=6,
        downsample=3,
        vals_to_return=["imgs-0"],
        sampling="all",
        debug=False,
    ):
        """
        Hierarchical Dataloader supporting fetching multiple views and multiple blocks of a video
        Can be extended to support additional modalities as desired
        :param mode: train, test
        :param transform: transforms to be applied to the data
        :param seq_len: Sequence length
        :param num_seq: Number of sequences
        :param downsample: Downsample rate
        :param vals_to_return: Which modalities to return
        :param sampling: Whether to return one block or all blocks from a video
        :param debug: Debugging mode, heavily reduces the amount of data being read for faster debugging
        """

        # Initialize basic info
        self.dataset = "homage"
        self.mode = mode
        self.debug = debug
        self.transform = transform
        self.seq_len = seq_len
        self.num_seq = num_seq
        self.downsample = downsample
        self.vals_to_return = set(vals_to_return)

        # sampling: all, single
        self.overlap = 0.25
        self.sampling = sampling
        self.sampler = self.sample_all_blocks if self.sampling == 'all' else self.sample_a_block

        self.num_classes = {'video': mu.get_num_classes('homage'), 'atomic': mu.get_num_classes('homage-atomic')}
        self.dense_labels, self.split, self.action_info, self.action_dict, self.row_info = \
            self.populate_video_annotations_details()
        self.video_names = list(self.dense_labels['video'].keys())

        random.Random(42).shuffle(self.video_names)

        self.video_names = self.video_names[::-1]

        # Randomly choose subsample as debug set
        if self.debug:
            self.video_names = self.video_names[:25]

        # Populate the number of frames in each video
        self.num_video_frames = {}

        for video in self.dense_labels['video'].keys():
            print(video)
            #video = video[3:]
            frame_ids = [int(f.replace('.png', '').split('_')[-1]) for f in os.listdir(video)]
            self.num_video_frames[video] = max(frame_ids)

    def get_action_idx(self, action_file):
        action_dict_decode, action_dict_encode = {}, {}
        action_df = pd.read_csv(action_file, sep=',', header=None)
        for _, row in action_df.iterrows():
            act_id, act_name = row
            act_id = int(act_id) - 1 # let id start from 0
            action_dict_decode[act_id] = act_name
            action_dict_encode[act_name] = act_id
        return action_dict_decode, action_dict_encode

    def fetch_imgs_seq(self, vpath, seq_len, idx_block):
        """
        Fetches the imgs sequence specified by the path and seq_idx
        """
        img_list = [os.path.join(vpath, '%06d.png' % (i + 1)) for i in idx_block]
        seq = [pil_loader(f) for f in img_list]
        img_t_seq = self.transform["imgs"](seq)  # apply same transform
        (IC, IH, IW) = img_t_seq[0].size()
        img_t_seq = torch.stack(img_t_seq, 0)
        img_t_seq = img_t_seq.view(self.num_seq, seq_len, IC, IH, IW).transpose(1, 2)
        return img_t_seq

    def fetch_audio_seq(self, vpath, idx_block):
        """
        Fetches the audio spectrogram sequence specified by the path and idx_block
        """

        # Load the spectrogram image
        person, vid = vpath.split('/')[-2:]
        spectrogram_path = '{}/data/homage/spectrogram/{}/{}_spec.jpg'.format(
            BASE_DIR, person, vid)

        # Get the cropped spectrogram
        minFrameIdx, maxFrameIdx = np.min(idx_block), np.max(idx_block)
        audioFrameLength = get_spectrogram_window_length(self.seq_len, self.num_seq, self.downsample)

        # Safely load image, otherwise pass zero tensor (We have around 20 audio files without spectrograms)
        if os.path.isfile(spectrogram_path):
            spectrogram = pil_loader(spectrogram_path)
        else:
            spectrogram = np.zeros((audioFrameLength + 1, 128, 3), dtype=np.float32)

        spectrogram = self.transform[mu.AudioMode](spectrogram)
        origShape = spectrogram.shape

        # Choose mid based on handling right edge case
        midAudioFrameIdx = int(imgFrameIdxToAudioIdxFactor * (minFrameIdx + maxFrameIdx) * 0.5)
        midAudioFrameIdx = min(midAudioFrameIdx, int(origShape[1] - (0.5 * audioFrameLength)))

        # Choose start based on handling left edge case. Start from 1 because negative indexing becomes an issue with 0
        startAudioIdx = max(1, int(midAudioFrameIdx - (0.5 * audioFrameLength)))

        # Note: We reverse the order of the spectrogram due to the nature it's dumped in
        cropped_spectrogram = spectrogram[:, -(startAudioIdx + audioFrameLength):-(startAudioIdx), ...]
        cropped_spectrogram = torch.flip(cropped_spectrogram, [1])

        assert cropped_spectrogram.shape[1] == audioFrameLength,\
            "Invalid shape: {}, Orig shape: {}, Misc: {}, {}, {}, {}, {}, Path: {}".format(
                cropped_spectrogram.shape, origShape, audioFrameLength,
                minFrameIdx, maxFrameIdx, startAudioIdx, midAudioFrameIdx, spectrogram_path)

        # Returned image is channels x time x 128
        return cropped_spectrogram

    def get_row_details(self, row):
        row = deepcopy(list(row))

        num_views = int(len([i for i in row if i == i]) / 4)
        i0 = [i for i in range(num_views) if 'v000' in row[i * 4]][0]
        i1 = np.random.choice(np.setdiff1d(range(num_views), [i0]))
        vpath0, vstart0, vend0, vname, vpath1, vstart1, vend1, _ = row[4*i0: 4*i0+4] + row[4*i1: 4*i1+4]

        return vpath0, vstart0, vend0, vname, vpath1, vstart1, vend1

    def populate_video_annotations_details(self):
        '''
        Populates video annotation details based on video-level and atomic-action level details
        Goes over each time step and populates the actions and the associated labels
        '''

        levels = ['video', 'atomic']

        dense_labels, split, action_info, action_dict, row_info = {}, {}, {}, {}, {}

        # Go over each level and perform the necessary loading operations for it
        for level in levels:
            postfix = '' if level == 'video' else '_atomic'

            # load the annotation csvs
            split[level] = os.path.join('{}/data/homage/{}_split{}.csv'.format(
                BASE_DIR, self.mode, postfix))
            action_info[level] = pd.read_csv(split[level], header=None, names=list(range(20)))

            # Get the action_dict encoders and decoders
            action_file = os.path.join('./data/' + self.dataset, 'classInd{}.txt'.format(postfix))
            action_dict[level] = {}
            action_dict[level]['decode'], action_dict[level]['encode'] = self.get_action_idx(action_file)

            # Store info about the row
            row_info[level] = {}

            # Populate all action segments for each video
            dense_labels[level] = {}
            for idx, row in tqdm(action_info[level].iterrows(), total=len(action_info[level])):
                # Get vname, list((start, end, (atomic)action_id))
                num_views = int(len([i for i in np.array(row) if i == i]) / 4)
                # Hacky way to find out view 0 idx
                v0_idx = next(iter([i for i in range(num_views) if 'v000' in row[i * 4]]), None)
                if v0_idx is None:
                    # We don't have an ego view
                    continue
                vpath, start, end, action = row[4*v0_idx: 4*v0_idx + 4]
                #print(vpath, start, end, action)
                if vpath not in dense_labels[level]:
                    dense_labels[level][vpath] = []
                dense_labels[level][vpath].append((int(start), int(end), action_dict[level]['encode'][action]))
                # Add info about the row
                row_info[level][vpath] = row
            #print(dense_labels)

        return dense_labels, split, action_info, action_dict, row_info

    def sample_all_blocks(self, seq_len, num_seq, video, numFrames=None, maxBlocks=50):
        '''
        sample all indices from a video
        Returns T blocks of num_seq x seq_len indices
        maxBlocks refers to the maximum number of segments from each video
        '''

        # Amount of overlap between the blocks
        overlap = self.overlap
        downsample = self.downsample
        totFrames = int(num_seq * seq_len * downsample)
        if numFrames is None:
            numFrames = self.num_video_frames[video]

        jump = 1 - overlap
        # Increase the jump if the number of blocks becomes too much
        if (numFrames / int(jump * totFrames)) >= maxBlocks:
            jump = (numFrames - 1) / (maxBlocks * totFrames * 1.0)

        sequences = []
        # Randomized initial start point from the first 10% of the video
        # start = random.randint(0, min(int(numFrames * 0.1), max(0, numFrames - totFrames - 1)))
        # Start from the first frame
        start = 0

        # Add dynamic sampling in case there are not enough frames
        if ((numFrames - start) - totFrames) <= 0:
            downsample = ((numFrames - start) / (num_seq * seq_len * 1.0)) * 0.95
            totFrames = int(num_seq * seq_len * downsample)

        while (start + totFrames) <= numFrames:
            sequence_block = np.array(np.arange(start, start + totFrames, downsample)).astype(int).reshape(num_seq, seq_len)
            sequences.append(sequence_block)
            start += int(jump * totFrames)

        sequences = np.array(sequences)

        return sequences

    def sample_a_block(self, seq_len, num_seq, video, numFrames=None):
        '''
        sample a block indices from a video
        Returns 1 block of num_seq x seq_len indices
        '''

        # Amount of overlap between the blocks
        downsample = self.downsample
        totFrames = int(num_seq * seq_len * downsample)
        if numFrames is None:
            numFrames = self.num_video_frames[video]

        sequences = []

        # Randomized initial start point from the video
        start = random.randint(0, max(0, numFrames - totFrames))

        # Add dynamic sampling in case there are not enough frames
        if ((numFrames - start) - totFrames) <= 0:
            downsample = ((numFrames - start) / (num_seq * seq_len * 1.0)) * 0.95
            totFrames = int(num_seq * seq_len * downsample)

        sequence_block = np.array(np.arange(start, start + totFrames, downsample)).astype(int).reshape(num_seq, seq_len)
        sequences.append(sequence_block)

        sequences = np.array(sequences)

        return sequences

    def __len__(self):
        return len(self.video_names)

    def get_labels_for_sequences(self, sequences, video):
        '''
        Returns video and atomic action labels
        sequences: B x num_seq x seq_len
        return B x 1, B x num_atomic_classes
        '''
        B = sequences.shape[0]

        atomic_labels_list = [[] for i in range(self.num_video_frames[video])]
        for segment in self.dense_labels['atomic'].get(video, []):
            start, end, action = segment
            for j in range(start, end):
                atomic_labels_list[j].append(action)
        video_labels = torch.tensor([self.dense_labels['video'][video][0][-1] for _ in range(B)])

        atomic_labels = torch.zeros((B, self.num_classes['atomic'])).long()
        for idx in range(sequences.shape[0]):
            for frame in range(sequences[idx][0, 0], sequences[idx][-1, -1]):
                for action in atomic_labels_list[frame]:
                    atomic_labels[idx][action] = 1

        return video_labels, atomic_labels

    def get_video_name(self, index):
        return self.video_names[index]

    def get_ego_third_view_details(self, index):
        ego_path = self.get_video_name(index)
        row = self.row_info['video'][ego_path]
        # Sample a 3rd person view as well
        vpath0, _, vend0, vname, vpath1, _, vend1 = self.get_row_details(row)
        vend = min(vend0, vend1)
        assert ego_path == vpath0, "Mismatch in the path {}, {}".format(ego_path, vpath0)
        return vpath1, vend

    def __getitem__(self, index):
        video = self.get_video_name(index)
        third_person_video, nframes = self.get_ego_third_view_details(index)

        vals = {}
        sequences = self.sampler(self.seq_len, self.num_seq, video, nframes)

        # Init val with lists
        #modes = ["imgs-0", "imgs-1", mu.AudioMode]
        modes = ["imgs-0", "imgs-1"]

        for val in modes:
            if val in self.vals_to_return:
                vals[val] = []

        # Go over all the sequences
        for sequence in sequences:
            idx_block = sequence.reshape(self.seq_len * self.num_seq)
            # 0 stands for the ego-view while 1 stands for the third-person view
            if "imgs-0" in self.vals_to_return:
                img_t_seq0 = self.fetch_imgs_seq(video, self.seq_len, idx_block)
                vals['{}-0'.format(mu.ImgMode)].append(img_t_seq0)
            if "imgs-1" in self.vals_to_return:
                img_t_seq1 = self.fetch_imgs_seq(third_person_video, self.seq_len, idx_block)
                vals['{}-1'.format(mu.ImgMode)].append(img_t_seq1)
            if mu.AudioMode in self.vals_to_return:
                vals[mu.AudioMode].append(self.fetch_audio_seq(video, idx_block))

        for val in modes:
            if val in self.vals_to_return:
                try:
                    vals[val] = torch.stack(vals[val])
                    if self.sampling == 'single':
                        # Squeeze the first singular dim
                        vals[val] = vals[val].squeeze(0)
                except:
                    print(sequences.shape, video)

        if "labels" in self.vals_to_return:
            vals["video_labels"], vals["atomic_labels"] = self.get_labels_for_sequences(sequences, video)
            if self.sampling == 'single':
                for val in ["video_labels", "atomic_labels"]:
                    # Squeeze the first singular dim
                    vals[val] = vals[val].squeeze(0)

        # Add video index field
        vals["vnames"] = torch.LongTensor([index])

        return vals


"""
Unit tests testing the functionality of the data loaders above
"""


import unittest


class TestHierarchicalHomage(unittest.TestCase):

    @classmethod
    def setUp(self):
        """
        This code code is ran once before all tests.
        """

        self.dataset = HierarchicalHomage(
            mode='val',
            transform=mu.get_test_transforms({"img_dim": 128}),
            seq_len=5,
            num_seq=8,
            downsample=3,
            #vals_to_return=["imgs-0", "audio", "labels"]
            vals_to_return=["imgs-0", "labels"]

        )

    def test_fetch_batch(self):
        for idx in range(25):
            batch = self.dataset[idx]
            B = batch['atomic_labels'].shape[0]
            vname = self.dataset.get_video_name(batch['vnames'][0]).split('/')[-1]
            print('Label counts:', vname, B, torch.unique(batch['atomic_labels'].sum(dim=-1), return_counts=True))


class TestHierarchicalHomageSingleSample(unittest.TestCase):

    @classmethod
    def setUp(self):
        """
        This code code is ran once before all tests.
        """

        self.dataset = HierarchicalHomage(
            mode='val',
            transform=mu.get_test_transforms({"img_dim": 128}),
            seq_len=5,
            num_seq=8,
            downsample=3,
            #vals_to_return=["imgs-0", "imgs-1", "audio", "labels"],
            vals_to_return=["imgs-0", "imgs-1", "labels"],

            sampling='single',
            debug=True
        )

    def test_fetch_batch(self):
        for idx in range(5):
            batch = self.dataset[idx]
            vname = self.dataset.get_video_name(batch['vnames'][0]).split('/')[-1]
            print([(k, v.shape) for k, v in batch.items()])
            print('Video label counts:', vname, torch.unique(batch['video_labels'], return_counts=True))
            print('Atomic counts:', vname, torch.unique(batch['atomic_labels'].sum(dim=-1), return_counts=True))


if __name__ == '__main__':
    unittest.main()
